from typing import Callable, List, Dict
from torch.utils.data.dataset import Dataset
from research_utils.sqlite.helper import Helper, FieldType
from research_utils.sqlite.sql import SQLArgument, SQLArgumentFactory

class DBFolder(Dataset):
  def __init__(
    self,
    db_path: str,
    table_name: str,
    fields: Dict[str, FieldType],
    argument: SQLArgument
  ) -> None:
    super().__init__()
    helper = Helper(
      db_path=db_path, 
      table_name=table_name,
      fields=fields
    )
    self.__rows = helper.query(
      fields=list(fields.keys()),
      argument=argument
    )
  
  def __len__(self):
    return len(self.__rows)
  
  def __getitem__(self, index):
    return self.__rows[index]


import math
import torch.nn as nn
from functools import reduce


def next_conv_dim(conv: nn.Conv2d, in_img_size: int):
  kernel_size = conv.kernel_size[0] if type(
    conv.kernel_size) is tuple else conv.kernel_size
  stride = conv.stride[0] if type(conv.stride) is tuple else conv.stride
  padding = conv.padding[0] if type(conv.padding) is tuple else conv.padding
  dilation = conv.dilation[0] if type(
    conv.dilation) is tuple else conv.dilation
  out_img_size = math.floor((
    in_img_size +
    2 * padding - dilation * (kernel_size - 1) - 1
  ) / stride + 1)
  return out_img_size

def next_conv_transpose_dim(convt: nn.ConvTranspose2d, in_img_size: int):
  kernel_size = convt.kernel_size[0] if type(
    convt.kernel_size) is tuple else convt.kernel_size
  stride = convt.stride[0] if type(convt.stride) is tuple else convt.stride
  padding = convt.padding[0] if type(convt.padding) is tuple else convt.padding
  dilation = convt.dilation[0] if type(
    convt.dilation) is tuple else convt.dilation
  output_padding = convt.output_padding[0] if type(convt.output_padding) is tuple else convt.output_padding

  out_img_size = math.floor(
    (in_img_size - 1) * stride - 
    2 * padding + dilation * (kernel_size - 1) + output_padding + 1
  )
  return out_img_size

def next_conv_block_dim(block: nn.Module, in_img_size: int):
  out_img_size = in_img_size
  for module in block:
    if isinstance(module, (nn.Conv2d, nn.MaxPool2d)):
      out_img_size = next_conv_dim(module, out_img_size)
    elif isinstance(module, nn.ConvTranspose2d):
      out_img_size = next_conv_transpose_dim(module, out_img_size)
    elif isinstance(module, (nn.Upsample, nn.UpsamplingBilinear2d, nn.UpsamplingNearest2d)):
      out_img_size *= module.scale_factor
  return out_img_size

def final_conv_dim(model: nn.Module, in_img_size: int):
  children = model.children()
  out_img_size = reduce(
    lambda _in_img_size, conv: next_conv_dim(conv, _in_img_size),
    children,
    in_img_size
  )
  return out_img_size


from typing import Dict, List
from enum import Enum
from functools import reduce
from sqlite_utils import Database
from .sql import SQLArgument

class FieldType(Enum):
  INTEGER = "INTEGER"
  FLOAT = "FLOAT"
  BOOL = "INTEGER"
  STRING = "TEXT"
  BLOB = "BLOB"


class Helper:
  def __init__(self, db_path, table_name, fields: Dict[str, FieldType], pk='id') -> None:
    self.__db = Database(db_path)
    self.__table_name = table_name
    self.__fields = fields

    if (not self.__db[self.__table_name].exists() and self.__fields is not None):
      self.__db[self.__table_name].create(self.__fields, pk=pk)
      print(f'created table: {self.__table_name}')
  
  @property
  def fields(self):
    return list(self.__fields.keys())
  @property
  def db(self):
    return self.__db

  def query(self, fields: List[str], argument: SQLArgument=None):
    select = reduce(lambda field1, field2: f'{field1}, {field2}', fields) if len(fields) > 0 else '*'
    sql = argument.sql if argument is not None else None
    print(select, sql)
    rows = self.__db[self.__table_name].rows_where(sql, select=select)
    return list(rows)


# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['research_utils',
 'research_utils.decorators',
 'research_utils.sqlite',
 'research_utils.torch']

package_data = \
{'': ['*']}

install_requires = \
['pytorch-lightning>=1.2.2,<2.0.0',
 'sqlite-utils>=3.6,<4.0',
 'torchvision>=0.8.2,<0.9.0']

setup_kwargs = {
    'name': 'research-utils',
    'version': '2021.9.18.1',
    'description': 'some utils for my research',
    'long_description': None,
    'author': 'Rainforest Cheng',
    'author_email': 'r08521610@ntu.edu.tw',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)
